//-------------------------------------------------------------------------------
//  TinyCircuits Tiny Arcade Menu
//
//  Tiny Arcade Menu is a TinyScreen+ program that runs after the TS+ bootloader
//  and looks for a connected SD card with TS+ binaries in the intended format.
//  If these files are found, the user is given a menu to select which binary to
//  write to the flash in a position after the menu, then jumps to that spot.
//  If no SD card is detected, jump to the last loaded application.
//  This program should be compiled and uploaded using Arduino SAMD Board support
//  1.6.7+ and TinyCircuits board package 1.0.5+  with the 'TinyArcade Menu' build
//  option.
//
//  Changelog:
//  1.0.2 3 October 2016 - Add compatibility for new Arcade revision, change
//        folder structure for IDE compatibility
//  1.0.1 16 May 2016 - Fix SD card detection timeout
//  1.0.0 Initial release
//
//  Written by Ben Rose for TinyCircuits, http://TinyCircuits.com
//
//-------------------------------------------------------------------------------

#include <TinyScreen.h>
#include "SdFat.h"
#include <Wire.h>
#include <SPI.h>
#include "MenuUtilAndDefs.h"
#include "TinyArcade.h"

TinyScreen display = TinyScreen(TinyScreenPlus);

SdFat sd;
SdFile dir;
SdFile file;
SdFile vidFile;
uint8_t buffer[96 * 64 * 2];

int currentFileNum = 0;
int doVideo = 0;
char currentGameName[50] = "  ";

int printNameOffset = 0;
int printNameDir = 0;
unsigned long lastPrintNameOffsetChange = 0;
unsigned long printNameOffsetChangeInterval = 200;


void setup(void) {
  arcadeInit();
  display.begin();
  display.setFont(thinPixel7_10ptFontInfo);
  printCentered("Finding SD card..", 46);
  display.setBitDepth(true);
  delay(20);
  for (int x = -37; x < 5; x++) {
    display.goTo(0, constrain(x, 0, 5));
    display.startData();
    int offset = (96 * constrain(-x, 0, 37) * 2);
    display.writeBuffer((uint8_t *)TinyCircuits96 + offset, (96 * 37 * 2) - offset);
    display.endTransfer();
    delay(20 + x / 3);
  }
  display.setBitDepth(false);
  delay(250 );
  if (!sd.begin(10, SPI_FULL_SPEED)) {
    display.clearScreen();
    printCentered("SD card not found,", 20);
    printCentered("looking for game..", 30);
    delay(500);
    if (*(uint32_t *)(APP_START_ADDRESS + 4) == 0xFFFFFFFF) {
      printCentered("No game loaded.", 40);
      printCentered("Upload or install SD", 40);
      USBDevice.init();
      USBDevice.attach();
      while (1);
    }
    SPI.end();
    SPI1.end();
    jumpApplication();
    while (1);
  }
  printNextFile();
}

int blockInput = 1;


void loop() {
  if (!checkJoystick(TAJoystickLeft) && !checkJoystick(TAJoystickRight)) {
    blockInput = 0;
  }
  if (!blockInput && checkJoystick(TAJoystickRight)) {
    printNextFile();
    blockInput = 1;
  }
  if (!blockInput && checkJoystick(TAJoystickLeft)) {
    printPreviousFile();
    blockInput = 1;
  }
  if (!blockInput && checkJoystick(TAJoystickDown)) {
    if ((checkButton(TAButton1) || checkButton(TAButton2)) && currentFileNum) {
      if (*(uint32_t *)(APP_START_ADDRESS + 4) != 0xFFFFFFFF) {
        SPI.end();
        SPI1.end();
        jumpApplication();
      }
    }
  }
  else if ((checkButton(TAButton1) || checkButton(TAButton2)) && currentFileNum) {
    for (int y = 0; y < 64; y++) {
      for (int x = 0; x < 96; x++) {
        ((uint16_t *)buffer)[(y * 96) + x] = (((uint16_t *)buffer)[(y * 96) + x] >> 1) & 0xEF7B; //0xE739
      }
    }
    writeToDisplay();
    writeFlash();
    SPI.end();
    SPI1.end();
    jumpApplication();
  }
  if (doVideo) {
    bufferVideoFrame();
    printName(currentGameName);
    writeToDisplay();
  } else {
    memset(buffer, 0, 96 * 64 * 2);
    printName(currentGameName);
    writeToDisplay();
  }
}

void printCentered(char * printString, int y) {
  display.setCursor(48 - ((int)display.getPrintWidth(printString) / 2), y);
  display.print(printString);
}

void bufferVideoFrame() {
  if (!vidFile.available())
    vidFile.rewind();
  vidFile.read(buffer, 96 * 64 * 2);
}

void writeToDisplay() {
  display.setBitDepth(1);
  display.goTo(0, 0);
  display.startData();
  display.writeBuffer(buffer, 96 * 64 * 2);
  display.endTransfer();
  display.setBitDepth(0);
}

bool isValidGameDir() {
  if (dir.isDir()) {
    char path[51] = " ";
    int dirLength = 0;
    dir.getName(path, 50);
    dirLength = strlen(path);
    path[dirLength] = '/';
    dir.getName(path + dirLength + 1, 50);
    strcpy(path + dirLength * 2 + 1, ".bin");
    file.close();
    if (file.open(path, O_READ)) {
      strcpy(path + dirLength * 2 + 1, ".tsv");
      if (vidFile.open(path, O_READ))
        doVideo = 1;
      return true;
    }
  }
  return false;
}

void printNextFile() {
  doVideo = 0;
  vidFile.close();
  int foundGame = 0;
  while (!foundGame) {
    dir.close();
    if (dir.openNext(sd.vwd(), O_READ)) {
      currentFileNum++;
    } else {
      sd.vwd()->rewind();
      dir.openNext(sd.vwd(), O_READ);
      currentFileNum = 1;
    }
    foundGame = isValidGameDir();
  }
  dir.getName(currentGameName, 50);
  printNameOffset = 0;
  printNameDir = 0;
  lastPrintNameOffsetChange = millis();
}

void printPreviousFile() {
  doVideo = 0;
  vidFile.close();
  int foundGame = 0;
  while (!foundGame) {
    if (currentFileNum > 1) {
      currentFileNum--;
    } else {
      currentFileNum = 100;
      int i;
      for (i = 1; i < currentFileNum; i++) {
        dir.close();
        if (!dir.openNext(sd.vwd(), O_READ))break;
      }
      currentFileNum = i;
    }
    sd.vwd()->rewind();
    int i;
    for (i = 1; i <= currentFileNum; i++) {
      dir.close();
      if (!dir.openNext(sd.vwd(), O_READ))break;
    }
    foundGame = isValidGameDir();
  }
  dir.getName(currentGameName, 50);
  printNameOffset = 0;
  printNameDir = 0;
  lastPrintNameOffsetChange = millis();
}

void printName(char * name) {
  int yOffset = 34;
  for (int y = yOffset + 0; y < yOffset + 12; y++) {
    for (int x = 0; x < 96; x++) {
      ((uint16_t *)buffer)[(y * 96) + x] = (((uint16_t *)buffer)[(y * 96) + x] >> 2) & 0xE739; //0xEF7B
    }
  }
  char displayName[50];
  strcpy(displayName, name + printNameOffset);

  int xPosition = 48 - ((int)display.getPrintWidth(displayName) / 2);

  if (millis() > lastPrintNameOffsetChange + printNameOffsetChangeInterval) {
    lastPrintNameOffsetChange = millis();
    if (xPosition < 9) {
      printNameOffset++;
    } else {
      printNameOffset = 0;
    }
  }

  while (xPosition < 9) {
    displayName[strlen(displayName) - 1] = '\0';
    xPosition = 48 - ((int)display.getPrintWidth(displayName) / 2);
  }

  for (int y = yOffset + 1; y < yOffset + 15; y++) {
    putString(y, xPosition, yOffset + 1, displayName, buffer + (96 * 2 * y), thinPixel7_10ptFontInfo);
    putString(y, 1, yOffset + 1, "<", buffer + (96 * 2 * y), thinPixel7_10ptFontInfo);
    putString(y, 87 + 5, yOffset + 1, ">", buffer + (96 * 2 * y), thinPixel7_10ptFontInfo);
  }
}

void putString(int y, int fontX, int fontY, char * string, uint8_t * buff, const FONT_INFO & fontInfo) {
  const FONT_CHAR_INFO* fontDescriptor = fontInfo.charDesc;
  int fontHeight = fontInfo.height;
  if (y >= fontY && y < fontY + fontHeight) {
    const unsigned char* fontBitmap = fontInfo.bitmap;
    int fontFirstCh = fontInfo.startCh;
    int fontLastCh = fontInfo.endCh;
    int stringChar = 0;
    int ch = string[stringChar++];
    while (ch) {
      uint8_t chWidth = pgm_read_byte(&fontDescriptor[ch - fontFirstCh].width);
      int bytesPerRow = chWidth / 8;
      if (chWidth > bytesPerRow * 8)
        bytesPerRow++;
      unsigned int offset = pgm_read_word(&fontDescriptor[ch - fontFirstCh].offset) + (bytesPerRow * fontHeight) - 1;

      for (uint8_t byte = 0; byte < bytesPerRow; byte++) {
        uint8_t data = pgm_read_byte(fontBitmap + offset - (y - fontY) - ((bytesPerRow - byte - 1) * fontHeight));
        uint8_t bits = byte * 8;
        for (int i = 0; i < 8 && (bits + i) < chWidth; i++) {
          if (data & (0x80 >> i)) {
            buff[(fontX) * 2] = 0xFFFF >> 8;
            buff[(fontX) * 2 + 1] = 0xFFFF;
          } else {

          }
          fontX++;
        }
      }
      fontX += 1;
      ch = string[stringChar++];
    }
  }
}

void showStatusBar(int state, unsigned int color1, unsigned int color2) {
  for (int y = 20; y < 24; y++) {
    for (int x = 0; x < 96; x++) {
      if (x < state)
        ((uint16_t *)buffer)[(y * 96) + x] = color1;
      else
        ((uint16_t *)buffer)[(y * 96) + x] = color2;
    }
  }
  display.setBitDepth(1);
  display.goTo(0, 20);
  display.startData();
  display.writeBuffer(buffer + (96 * 2) * 20, (96 * 2) * 4);
  display.endTransfer();
  display.setBitDepth(0);
}

void writeFlash() {
  uint32_t address = 64 * 1024;

  uint32_t PAGE_SIZE, PAGES, MAX_FLASH;
  uint32_t pageSizes[] = { 8, 16, 32, 64, 128, 256, 512, 1024 };
  PAGE_SIZE = pageSizes[NVMCTRL->PARAM.bit.PSZ];
  PAGES = NVMCTRL->PARAM.bit.NVMP;
  MAX_FLASH = PAGE_SIZE * PAGES;

  uint32_t erase_dst_addr = address; // starting address

  int lastDisplayedValue = -1;

  while (erase_dst_addr < MAX_FLASH) {
    // Execute "ER" Erase Row
    NVMCTRL->ADDR.reg = erase_dst_addr / 2;
    NVMCTRL->CTRLA.reg = NVMCTRL_CTRLA_CMDEX_KEY | NVMCTRL_CTRLA_CMD_ER;

    uint32_t newDisplayValue = ((erase_dst_addr - address) * 96) / (MAX_FLASH - address);
    if ( newDisplayValue != lastDisplayedValue) {
      showStatusBar(newDisplayValue, 0x1F00, 0x0000);
      lastDisplayedValue = newDisplayValue;
    }


    while (NVMCTRL->INTFLAG.bit.READY == 0)
      ;
    erase_dst_addr += PAGE_SIZE * 4; // Skip a ROW
  }

  lastDisplayedValue = -1;

  // Write to flash
  uint32_t size = file.fileSize(); // PAGE_SIZE;
  uint32_t *dst_addr = (uint32_t *)(64 * 1024);
  uint8_t *ptr_data = buffer;

  // Set automatic page write
  NVMCTRL->CTRLB.bit.MANW = 0;
  // Set first address
  NVMCTRL->ADDR.reg = address / 2;
  // Do writes in pages
  while (size) {


    int amtBytes = file.read(buffer, PAGE_SIZE);

    if (size < 64) {
      for (int j = size; j < PAGE_SIZE; j++)buffer[j] = 0xFF;
    }
    // Execute "PBC" Page Buffer Clear
    NVMCTRL->CTRLA.reg = NVMCTRL_CTRLA_CMDEX_KEY | NVMCTRL_CTRLA_CMD_PBC;
    while (NVMCTRL->INTFLAG.bit.READY == 0);

    // Fill page buffer
    uint32_t i;
    for (i = 0; i < (PAGE_SIZE / 4) && i < size; i++) {
      //uint32_t data = (buffer[i + 3] << 0) | (buffer[i + 2] << 1) | (buffer[i + 1] << 2) | (buffer[i] << 3);
      dst_addr[i] = ((uint32_t *)buffer)[i];
    }

    // Execute "WP" Write Page
    NVMCTRL->CTRLA.reg = NVMCTRL_CTRLA_CMDEX_KEY | NVMCTRL_CTRLA_CMD_WP;

    int newDisplayValue = 96 - (size * 96) / file.fileSize();
    if ( newDisplayValue != lastDisplayedValue) {
      showStatusBar(newDisplayValue, 0xE007, 0x1F00);
      lastDisplayedValue = newDisplayValue;
    }

    while (NVMCTRL->INTFLAG.bit.READY == 0);

    // Advance to next page
    dst_addr += i;
    if (size > PAGE_SIZE)
      size     -= PAGE_SIZE;
    else
      size = 0;
  }
}
